@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2022 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -56,6 +56,9 @@ NOEXPORT void local_start(CLI *);
 NOEXPORT void remote_start(CLI *);
 NOEXPORT void ssl_start(CLI *);
 NOEXPORT void session_cache_retrieve(CLI *);
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+NOEXPORT void print_tmp_key(SSL *s);
+#endif
 NOEXPORT void print_cipher(CLI *);
 NOEXPORT void transfer(CLI *);
 NOEXPORT int parse_socket_error(CLI *, const char *);
@@ -94,12 +97,28 @@ unsigned __stdcall
 #else
 void *
 #endif
-client_thread(void *arg) {
+        client_thread(void *arg) {
     CLI *c=arg;
 #ifdef DEBUG_STACK_SIZE
     size_t stack_size=c->opt->stack_size;
 #endif
 
+#ifdef USE_FORK
+    /* do not use signal pipe in child processes */
+    signal(SIGCHLD, SIG_IGN); /* ignore dead children */
+    signal(SIGHUP, SIG_DFL);
+    signal(SIGUSR1, SIG_DFL);
+    signal(SIGUSR2, SIG_DFL);
+    signal(SIGPIPE, SIG_IGN); /* ignore broken pipe */
+    signal(SIGTERM, SIG_DFL);
+    signal(SIGQUIT, SIG_DFL);
+    signal(SIGINT, SIG_DFL);
+#endif /* USE_FORK */
+
+    /* make sure c->thread_* values are initialized */
+    CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_THREAD_LIST]);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
+
     /* initialize */
     c->tls=NULL; /* do not reuse */
     tls_alloc(c, NULL, NULL);
@@ -110,7 +129,24 @@ client_thread(void *arg) {
     /* execute */
     client_main(c);
 
-    /* cleanup */
+    /* cleanup the thread */
+#ifndef USE_FORK
+    CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_THREAD_LIST]);
+    if(thread_head==c)
+        thread_head=c->thread_next;
+    if(c->thread_prev)
+        c->thread_prev->thread_next=c->thread_next;
+    if(c->thread_next)
+        c->thread_next->thread_prev=c->thread_prev;
+#ifdef USE_PTHREAD
+    pthread_detach(c->thread_id);
+#endif
+#ifdef USE_WIN32
+    CloseHandle(c->thread_id);
+#endif
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
+#endif /* !USE_FORK */
+    client_free(c);
 #ifdef DEBUG_STACK_SIZE
     stack_info(stack_size, 0); /* display computed value */
 #endif
@@ -118,7 +154,7 @@ client_thread(void *arg) {
     tls_cleanup();
     /* s_log() is not allowed after tls_cleanup() */
 
-    /* terminate */
+    /* terminate the thread */
 #if defined(USE_WIN32) || defined(USE_OS2)
 #if !defined(_WIN32_WCE)
     _endthreadex(0);
@@ -148,6 +184,9 @@ void client_main(CLI *c) {
     } else {
         client_run(c);
     }
+}
+
+void client_free(CLI *c) {
 #ifndef USE_FORK
     service_free(c->opt);
 #endif
@@ -180,7 +219,7 @@ NOEXPORT void exec_connect_loop(CLI *c) {
             s_log(LOG_INFO, "Retrying an exec+connect section");
             /* c and id are detached, so it is safe to call str_stats() */
             str_stats(); /* client thread allocation tracking */
-            sleep(1); /* FIXME: not a good idea in ucontext threading */
+            s_poll_sleep(1, 0);
             c->rr++;
         }
 
@@ -268,6 +307,10 @@ NOEXPORT void client_run(CLI *c) {
         "Connection %s: %llu byte(s) sent to TLS, %llu byte(s) sent to socket",
         rst ? "reset" : "closed",
         (unsigned long long)c->ssl_bytes, (unsigned long long)c->sock_bytes);
+#ifdef USE_WIN32
+    if(capwin_hwnd && err==1 && InterlockedExchange(&capwin_connectivity, 0))
+        PostMessage(capwin_hwnd, WM_CAPWIN_NET_DOWN, 0, 0);
+#endif
 
         /* cleanup temporary (e.g. IDENT) socket */
     if(c->fd!=INVALID_SOCKET)
@@ -374,15 +417,19 @@ NOEXPORT void local_start(CLI *c) {
     addr_len=sizeof(SOCKADDR_UNION);
     c->local_rfd.is_socket=!getpeername(c->local_rfd.fd, &addr.sa, &addr_len);
     if(c->local_rfd.is_socket) {
+        /* store the retrieved peer address */
         memcpy(&c->peer_addr.sa, &addr.sa, (size_t)addr_len);
         c->peer_addr_len=addr_len;
-        if(socket_options_set(c->opt, c->local_rfd.fd, 1))
+        if(
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+                addr.sa.sa_family!=AF_UNIX &&
+#endif
+                socket_options_set(c->opt, c->local_rfd.fd, 1)) {
             s_log(LOG_WARNING, "Failed to set local socket options");
-    } else {
-        if(get_last_socket_error()!=S_ENOTSOCK) {
-            sockerror("getpeerbyname (local_rfd)");
-            throw_exception(c, 1);
         }
+    } else if(get_last_socket_error()!=S_ENOTSOCK) {
+        sockerror("getpeerbyname (local_rfd)");
+        throw_exception(c, 1);
     }
 
     /* check if local_wfd is a socket and get peer address */
@@ -392,17 +439,21 @@ NOEXPORT void local_start(CLI *c) {
         addr_len=sizeof(SOCKADDR_UNION);
         c->local_wfd.is_socket=!getpeername(c->local_wfd.fd, &addr.sa, &addr_len);
         if(c->local_wfd.is_socket) {
-            if(!c->local_rfd.is_socket) { /* already retrieved */
+            if(!c->local_rfd.is_socket) { /* not retrieved from rfd? */
+                /* store the retrieved peer address */
                 memcpy(&c->peer_addr.sa, &addr.sa, (size_t)addr_len);
                 c->peer_addr_len=addr_len;
             }
-            if(socket_options_set(c->opt, c->local_wfd.fd, 1))
+            if(
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+                    addr.sa.sa_family!=AF_UNIX &&
+#endif
+                    socket_options_set(c->opt, c->local_wfd.fd, 1)) {
                 s_log(LOG_WARNING, "Failed to set local socket options");
-        } else {
-            if(get_last_socket_error()!=S_ENOTSOCK) {
-                sockerror("getpeerbyname (local_wfd)");
-                throw_exception(c, 1);
             }
+        } else if(get_last_socket_error()!=S_ENOTSOCK) {
+            sockerror("getpeerbyname (local_wfd)");
+            throw_exception(c, 1);
         }
     }
 
@@ -461,7 +512,10 @@ NOEXPORT void remote_start(CLI *c) {
 
 NOEXPORT void ssl_start(CLI *c) {
     int i, err;
+    SSL_SESSION *sess;
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     int unsafe_openssl;
+#endif /* OpenSSL version < 1.1.0 */
 
     c->ssl=SSL_new(c->opt->ctx);
     if(!c->ssl) {
@@ -507,31 +561,34 @@ NOEXPORT void ssl_start(CLI *c) {
     else
         s_log(LOG_INFO, "Peer certificate not required");
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     unsafe_openssl=OpenSSL_version_num()<0x0090810fL ||
         (OpenSSL_version_num()>=0x10000000L &&
         OpenSSL_version_num()<0x1000002fL);
+#endif /* OpenSSL version < 1.1.0 */
     while(1) {
         /* critical section for OpenSSL version < 0.9.8p or 1.x.x < 1.0.0b *
          * this critical section is a crude workaround for CVE-2010-3864   *
          * see http://www.securityfocus.com/bid/44884 for details          *
          * alternative solution is to disable internal session caching     *
          * NOTE: this critical section also covers callbacks (e.g. OCSP)   */
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
         if(unsafe_openssl)
             CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_SSL]);
+#endif /* OpenSSL version < 1.1.0 */
 
-        if(c->opt->option.client)
-            i=SSL_connect(c->ssl);
-        else
-            i=SSL_accept(c->ssl);
+        i=c->opt->option.client ? SSL_connect(c->ssl) : SSL_accept(c->ssl);
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
         if(unsafe_openssl)
             CRYPTO_THREAD_unlock(stunnel_locks[LOCK_SSL]);
+#endif /* OpenSSL version < 1.1.0 */
 
         err=SSL_get_error(c->ssl, i);
         if(err==SSL_ERROR_NONE)
             break; /* ok -> done */
         if(err==SSL_ERROR_WANT_READ || err==SSL_ERROR_WANT_WRITE) {
-            s_poll_init(c->fds);
+            s_poll_init(c->fds, 0);
             s_poll_add(c->fds, c->ssl_rfd->fd,
                 err==SSL_ERROR_WANT_READ,
                 err==SSL_ERROR_WANT_WRITE);
@@ -561,16 +618,38 @@ NOEXPORT void ssl_start(CLI *c) {
 #endif
                 continue;
             }
+            sockerror(c->opt->option.client ? "SSL_connect" : "SSL_accept");
+            throw_exception(c, 1);
         }
-        if(c->opt->option.client)
-            sslerror("SSL_connect");
-        else
-            sslerror("SSL_accept");
+        sslerror(c->opt->option.client ? "SSL_connect" : "SSL_accept");
         throw_exception(c, 1);
     }
     print_cipher(c);
-    if(SSL_session_reused(c->ssl))
-        print_session_id(SSL_get_session(c->ssl));
+    sess=SSL_get1_session(c->ssl);
+    if(sess) {
+        X509 *peer_cert=SSL_get_peer_certificate(c->ssl);
+        if(peer_cert) {
+            X509_free(peer_cert);
+        } else { /* no authentication was performed */
+            if(!SSL_SESSION_set_ex_data(sess,
+                    index_session_authenticated, NULL)) {
+                sslerror("SSL_SESSION_set_ex_data");
+                SSL_SESSION_free(sess);
+                throw_exception(c, 1);
+            }
+        }
+        if(SSL_session_reused(c->ssl)) {
+            print_session_id(sess);
+        } else { /* a new session was negotiated */
+            /* SSL_SESS_CACHE_NO_INTERNAL_STORE prevented automatic caching */
+            if(!c->opt->option.client)
+                SSL_CTX_add_session(c->opt->ctx, sess);
+        }
+        SSL_SESSION_free(sess);
+    } else if(c->opt->redirect_addr.names) {
+        s_log(LOG_ERR, "No session available for redirection");
+        throw_exception(c, 1);
+    }
 }
 
 NOEXPORT void session_cache_retrieve(CLI *c) {
@@ -592,6 +671,49 @@ NOEXPORT void session_cache_retrieve(CLI *c) {
     CRYPTO_THREAD_unlock(stunnel_locks[LOCK_SESSION]);
 }
 
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+NOEXPORT void print_tmp_key(SSL *s) {
+    EVP_PKEY *key;
+
+#ifdef SSL_CTRL_GET_PEER_TMP_KEY
+    if (!SSL_get_peer_tmp_key(s, &key)) {
+        sslerror("SSL_get_peer_tmp_key");
+        return;
+    }
+#else
+    if (!SSL_get_server_tmp_key(s, &key)) {
+        sslerror("SSL_get_server_tmp_key");
+        return;
+    }
+#endif
+    switch (EVP_PKEY_id(key)) {
+    case EVP_PKEY_RSA:
+        s_log(LOG_INFO, "Peer temporary key: RSA, %d bits", EVP_PKEY_bits(key));
+        break;
+    case EVP_PKEY_DH:
+        s_log(LOG_INFO, "Peer temporary key: DH, %d bits", EVP_PKEY_bits(key));
+        break;
+#ifndef OPENSSL_NO_EC
+    case EVP_PKEY_EC:
+        {
+            EC_KEY *ec=EVP_PKEY_get1_EC_KEY(key);
+            int nid=EC_GROUP_get_curve_name(EC_KEY_get0_group(ec));
+            const char *cname=EC_curve_nid2nist(nid);
+            EC_KEY_free(ec);
+            if (cname == NULL)
+                cname=OBJ_nid2sn(nid);
+            s_log(LOG_INFO, "Peer temporary key: ECDH, %s, %d bits", cname, EVP_PKEY_bits(key));
+        }
+        break;
+#endif
+    default:
+        s_log(LOG_INFO, "Peer temporary key: %s, %d bits", OBJ_nid2sn(EVP_PKEY_id(key)),
+                   EVP_PKEY_bits(key));
+    }
+    EVP_PKEY_free(key);
+}
+#endif /* OpenSSL 1.1.1 or later */
+
 NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
     SSL_CIPHER *cipher;
 #ifndef OPENSSL_NO_COMP
@@ -610,6 +732,9 @@ NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
     s_log(LOG_INFO, "%s ciphersuite: %s (%d-bit encryption)",
         SSL_get_version(c->ssl), SSL_CIPHER_get_name(cipher),
         SSL_CIPHER_get_bits(cipher, NULL));
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+    print_tmp_key(c->ssl);
+#endif
 
 #ifndef OPENSSL_NO_COMP
     compression=SSL_get_current_compression(c->ssl);
@@ -623,8 +748,12 @@ NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
 
 /****************************** transfer data */
 NOEXPORT void transfer(CLI *c) {
+    int timeout; /* s_poll_wait timeout in seconds */
+    int pending; /* either processed on unprocessed TLS data */
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+    int has_pending=0, prev_has_pending;
+#endif
     int watchdog=0; /* a counter to detect an infinite loop */
-    ssize_t num;
     int err;
     /* logical channels (not file descriptors!) open for read or write */
     int sock_open_rd=1, sock_open_wr=1;
@@ -650,7 +779,7 @@ NOEXPORT void transfer(CLI *c) {
             && c->sock_ptr && !write_wants_read;
 
         /****************************** setup c->fds structure */
-        s_poll_init(c->fds); /* initialize the structure */
+        s_poll_init(c->fds, 0); /* initialize the structure */
         /* for plain socket open data strem = open file descriptor */
         /* make sure to add each open socket to receive exceptions! */
         if(sock_open_rd) /* only poll if the read file descriptor is open */
@@ -667,17 +796,31 @@ NOEXPORT void transfer(CLI *c) {
         }
 
         /****************************** wait for an event */
-        err=s_poll_wait(c->fds,
-            (sock_open_rd && /* both peers open */
+        pending=SSL_pending(c->ssl);
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+        /* only attempt to process SSL_has_pending() data once */
+        prev_has_pending=has_pending;
+        has_pending=SSL_has_pending(c->ssl);
+        pending=pending || (has_pending && !prev_has_pending);
+#endif
+        if(read_wants_read && pending) {
+            timeout=0; /* process any buffered data without delay */
+        } else if((sock_open_rd && /* both peers open */
                 !(SSL_get_shutdown(c->ssl)&SSL_RECEIVED_SHUTDOWN)) ||
-            c->ssl_ptr /* data buffered to write to socket */ ||
-            c->sock_ptr /* data buffered to write to TLS */ ?
-            c->opt->timeout_idle : c->opt->timeout_close, 0);
+                c->ssl_ptr /* data buffered to write to socket */ ||
+                c->sock_ptr /* data buffered to write to TLS */) {
+            timeout=c->opt->timeout_idle;
+        } else {
+            timeout=c->opt->timeout_close;
+        }
+        err=s_poll_wait(c->fds, timeout, 0);
         switch(err) {
         case -1:
             sockerror("transfer: s_poll_wait");
             throw_exception(c, 1);
         case 0: /* timeout */
+            if(read_wants_read && pending)
+                break;
             if((sock_open_rd &&
                     !(SSL_get_shutdown(c->ssl)&SSL_RECEIVED_SHUTDOWN)) ||
                     c->ssl_ptr || c->sock_ptr) {
@@ -685,12 +828,12 @@ NOEXPORT void transfer(CLI *c) {
                     " TIMEOUTidle exceeded: sending reset");
                 s_poll_dump(c->fds, LOG_DEBUG);
                 throw_exception(c, 1);
-            } else { /* already closing connection */
-                s_log(LOG_ERR, "transfer: s_poll_wait:"
-                    " TIMEOUTclose exceeded: closing");
-                s_poll_dump(c->fds, LOG_DEBUG);
-                return; /* OK */
             }
+            /* already closing connection */
+            s_log(LOG_ERR, "transfer: s_poll_wait:"
+                " TIMEOUTclose exceeded: closing");
+            s_poll_dump(c->fds, LOG_DEBUG);
+            return; /* OK */
         }
 
         /****************************** retrieve results from c->fds */
@@ -742,17 +885,21 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** hangups without read or write */
         if(!(sock_can_rd || sock_can_wr || ssl_can_rd || ssl_can_wr)) {
-            if(s_poll_hup(c->fds, c->sock_rfd->fd) ||
-                    s_poll_hup(c->fds, c->sock_wfd->fd)) {
+            if(s_poll_hup(c->fds, c->sock_wfd->fd)) {
                 if(c->ssl_ptr) {
                     s_log(LOG_ERR,
-                        "Socket closed (HUP) with %ld unsent byte(s)",
+                        "Write socket closed (HUP) with %ld unsent byte(s)",
                         (long)c->ssl_ptr);
                     throw_exception(c, 1); /* reset the sockets */
                 }
-                s_log(LOG_INFO, "Socket closed (HUP)");
-                sock_open_rd=sock_open_wr=0;
-            } else if(s_poll_hup(c->fds, c->ssl_rfd->fd) ||
+                s_log(LOG_INFO, "Write socket closed (HUP)");
+                sock_open_wr=0;
+            }
+            if(s_poll_hup(c->fds, c->sock_rfd->fd)) {
+                s_log(LOG_INFO, "Read socket closed (HUP)");
+                sock_open_rd=0;
+            }
+            if(s_poll_hup(c->fds, c->ssl_rfd->fd) ||
                     s_poll_hup(c->fds, c->ssl_wfd->fd)) {
                 if(c->sock_ptr) {
                     s_log(LOG_ERR,
@@ -773,9 +920,9 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** send TLS close_notify alert */
         if(shutdown_wants_read || shutdown_wants_write) {
-            num=SSL_shutdown(c->ssl); /* send close_notify alert */
+            int num=SSL_shutdown(c->ssl); /* send close_notify alert */
             if(num<0) /* -1 - not completed */
-                err=SSL_get_error(c->ssl, (int)num);
+                err=SSL_get_error(c->ssl, num);
             else /* 0 or 1 - success */
                 err=SSL_ERROR_NONE;
             switch(err) {
@@ -783,16 +930,6 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_INFO, "SSL_shutdown successfully sent close_notify alert");
                 shutdown_wants_read=shutdown_wants_write=0;
                 break;
-            case SSL_ERROR_SYSCALL: /* socket error */
-                if(parse_socket_error(c, "SSL_shutdown"))
-                    break; /* a non-critical error: retry */
-                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                shutdown_wants_read=shutdown_wants_write=0;
-                break;
-            case SSL_ERROR_ZERO_RETURN: /* connection closed */
-                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                shutdown_wants_read=shutdown_wants_write=0;
-                break;
             case SSL_ERROR_WANT_WRITE:
                 s_log(LOG_DEBUG, "SSL_shutdown returned WANT_WRITE: retrying");
                 shutdown_wants_read=0;
@@ -806,6 +943,16 @@ NOEXPORT void transfer(CLI *c) {
             case SSL_ERROR_SSL: /* TLS error */
                 sslerror("SSL_shutdown");
                 throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* received a close_notify alert */
+                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                shutdown_wants_read=shutdown_wants_write=0;
+                break;
+            case SSL_ERROR_SYSCALL: /* socket error */
+                if(parse_socket_error(c, "SSL_shutdown"))
+                    break; /* a non-critical error: retry */
+                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                shutdown_wants_read=shutdown_wants_write=0;
+                break;
             default:
                 s_log(LOG_ERR, "SSL_shutdown/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -814,7 +961,7 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** write to socket */
         if(sock_open_wr && sock_can_wr) {
-            num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
+            ssize_t num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
             switch(num) {
             case -1: /* error */
                 if(parse_socket_error(c, "writesocket"))
@@ -835,7 +982,7 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** read from socket */
         if(sock_open_rd && sock_can_rd) {
-            num=readsocket(c->sock_rfd->fd,
+            ssize_t num=readsocket(c->sock_rfd->fd,
                 c->sock_buff+c->sock_ptr, BUFFSIZE-c->sock_ptr);
             switch(num) {
             case -1:
@@ -863,10 +1010,10 @@ NOEXPORT void transfer(CLI *c) {
         /****************************** write to TLS */
         if((write_wants_read && ssl_can_rd) ||
                 (write_wants_write && ssl_can_wr)) {
+            int num=SSL_write(c->ssl, c->sock_buff, (int)(c->sock_ptr));
             write_wants_read=0;
             write_wants_write=0;
-            num=SSL_write(c->ssl, c->sock_buff, (int)(c->sock_ptr));
-            switch(err=SSL_get_error(c->ssl, (int)num)) {
+            switch(err=SSL_get_error(c->ssl, num)) {
             case SSL_ERROR_NONE:
                 if(num==0) { /* nothing was written: ignore */
                     s_log(LOG_DEBUG, "SSL_write returned 0");
@@ -891,8 +1038,13 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_DEBUG,
                     "SSL_write returned WANT_X509_LOOKUP: retrying");
                 break;
+            case SSL_ERROR_SSL:
+                sslerror("SSL_write");
+                throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* a buffered close_notify alert */
+                /* fall through */
             case SSL_ERROR_SYSCALL: /* socket error */
-                if(num && parse_socket_error(c, "SSL_write"))
+                if(parse_socket_error(c, "SSL_write") && num) /* always log the error */
                     break; /* a non-critical error: retry */
                 /* EOF -> buggy (e.g. Microsoft) peer:
                  * TLS socket closed without close_notify alert */
@@ -905,14 +1057,6 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_INFO, "TLS socket closed (SSL_write)");
                 SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                 break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify alert received */
-                s_log(LOG_INFO, "TLS closed (SSL_write)");
-                if(SSL_version(c->ssl)==SSL2_VERSION)
-                    SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                break;
-            case SSL_ERROR_SSL:
-                sslerror("SSL_write");
-                throw_exception(c, 1);
             default:
                 s_log(LOG_ERR, "SSL_write/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -920,14 +1064,14 @@ NOEXPORT void transfer(CLI *c) {
         }
 
         /****************************** read from TLS */
-        if((read_wants_read && (ssl_can_rd || SSL_pending(c->ssl))) ||
+        if((read_wants_read && (ssl_can_rd || pending)) ||
                 /* it may be possible to read some pending data after
                  * writesocket() above made some room in c->ssl_buff */
                 (read_wants_write && ssl_can_wr)) {
+            int num=SSL_read(c->ssl, c->ssl_buff+c->ssl_ptr, (int)(BUFFSIZE-c->ssl_ptr));
             read_wants_read=0;
             read_wants_write=0;
-            num=SSL_read(c->ssl, c->ssl_buff+c->ssl_ptr, (int)(BUFFSIZE-c->ssl_ptr));
-            switch(err=SSL_get_error(c->ssl, (int)num)) {
+            switch(err=SSL_get_error(c->ssl, num)) {
             case SSL_ERROR_NONE:
                 if(num==0) { /* nothing was read: ignore */
                     s_log(LOG_DEBUG, "SSL_read returned 0");
@@ -950,8 +1094,17 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_DEBUG,
                     "SSL_read returned WANT_X509_LOOKUP: retrying");
                 break;
+            case SSL_ERROR_SSL:
+                sslerror("SSL_read");
+                throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* received a close_notify alert */
+                s_log(LOG_INFO, "TLS closed (SSL_read)");
+                if(SSL_version(c->ssl)==SSL2_VERSION)
+                    SSL_set_shutdown(c->ssl,
+                        SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                break;
             case SSL_ERROR_SYSCALL:
-                if(num && parse_socket_error(c, "SSL_read"))
+                if(parse_socket_error(c, "SSL_read") && num) /* always log the error */
                     break; /* a non-critical error: retry */
                 /* EOF -> buggy (e.g. Microsoft) peer:
                  * TLS socket closed without close_notify alert */
@@ -965,15 +1118,6 @@ NOEXPORT void transfer(CLI *c) {
                 SSL_set_shutdown(c->ssl,
                     SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                 break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify alert received */
-                s_log(LOG_INFO, "TLS closed (SSL_read)");
-                if(SSL_version(c->ssl)==SSL2_VERSION)
-                    SSL_set_shutdown(c->ssl,
-                        SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                break;
-            case SSL_ERROR_SSL:
-                sslerror("SSL_read");
-                throw_exception(c, 1);
             default:
                 s_log(LOG_ERR, "SSL_read/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -1058,8 +1202,14 @@ NOEXPORT void transfer(CLI *c) {
             s_log(LOG_ERR,
                 "please report the problem to Michal.Trojnara@stunnel.org");
             stunnel_info(LOG_ERR);
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+            s_log(LOG_ERR, "protocol=%s, SSL_pending=%d, SSL_has_pending=%d",
+                SSL_get_version(c->ssl),
+                SSL_pending(c->ssl), SSL_has_pending(c->ssl));
+#else
             s_log(LOG_ERR, "protocol=%s, SSL_pending=%d",
                 SSL_get_version(c->ssl), SSL_pending(c->ssl));
+#endif
             s_log(LOG_ERR, "sock_open_rd=%s, sock_open_wr=%s",
                 sock_open_rd ? "Y" : "n", sock_open_wr ? "Y" : "n");
             s_log(LOG_ERR, "SSL_RECEIVED_SHUTDOWN=%s, SSL_SENT_SHUTDOWN=%s",
@@ -1086,22 +1236,15 @@ NOEXPORT void transfer(CLI *c) {
         shutdown_wants_read || shutdown_wants_write);
 }
 
-#ifdef __GNUC__
-#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
-#pragma GCC diagnostic push
-#endif /* __GNUC__>=4.6 */
-#if __GNUC__ >= 7
-#pragma GCC diagnostic ignored "-Wimplicit-fallthrough"
-#endif /* __GNUC__>=7 */
-#endif /* __GNUC__ */
-
     /* returns 0 on close and 1 on non-critical errors */
 NOEXPORT int parse_socket_error(CLI *c, const char *text) {
     switch(get_last_socket_error()) {
         /* http://tangentsoft.net/wskfaq/articles/bsd-compatibility.html */
     case 0: /* close on read, or close on write on WIN32 */
+        /* fall through */
 #ifndef USE_WIN32
     case EPIPE: /* close on write on Unix */
+        /* fall through */
 #endif
     case S_ECONNABORTED:
         s_log(LOG_INFO, "%s: Socket is closed", text);
@@ -1111,7 +1254,7 @@ NOEXPORT int parse_socket_error(CLI *c, const char *text) {
         return 1;
     case S_EWOULDBLOCK:
         s_log(LOG_NOTICE, "%s: Would block: retrying", text);
-        sleep(1); /* Microsoft bug KB177346 */
+        s_poll_sleep(1, 0); /* Microsoft bug KB177346 */
         return 1;
 #if S_EAGAIN!=S_EWOULDBLOCK
     case S_EAGAIN:
@@ -1126,8 +1269,8 @@ NOEXPORT int parse_socket_error(CLI *c, const char *text) {
             s_log(LOG_INFO, "%s: Socket is closed (exec)", text);
             return 0;
         }
-        /* fall through */
 #endif
+        /* fall through */
     default:
         sockerror(text);
         throw_exception(c, 1);
@@ -1135,12 +1278,6 @@ NOEXPORT int parse_socket_error(CLI *c, const char *text) {
     }
 }
 
-#ifdef __GNUC__
-#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
-#pragma GCC diagnostic pop
-#endif /* __GNUC__>=4.6 */
-#endif /* __GNUC__ */
-
 NOEXPORT void auth_user(CLI *c) {
 #ifndef _WIN32_WCE
     struct servent *s_ent;    /* structure for getservbyname */
@@ -1197,7 +1334,7 @@ NOEXPORT void auth_user(CLI *c) {
     }
     *system++='\0';
     if(strcmp(type, " USERID ")) {
-        s_log(LOG_ERR, "Incorrect INETD response type");
+        s_log(LOG_ERR, "Incorrect IDENT response type");
         str_free(line);
         throw_exception(c, 1);
     }
@@ -1306,6 +1443,7 @@ NOEXPORT SOCKET connect_local(CLI *c) { /* spawn local process */
         signal(SIGCHLD, SIG_DFL);
         signal(SIGHUP, SIG_DFL);
         signal(SIGUSR1, SIG_DFL);
+        signal(SIGUSR2, SIG_DFL);
         signal(SIGPIPE, SIG_DFL);
         signal(SIGTERM, SIG_DFL);
         signal(SIGQUIT, SIG_DFL);
@@ -1408,9 +1546,13 @@ NOEXPORT SOCKET connect_remote(CLI *c) {
         if(!connect_init(c, c->connect_addr.addr[c->idx].sa.sa_family) &&
                 !s_connect(c, &c->connect_addr.addr[c->idx],
                     addr_len(&c->connect_addr.addr[c->idx]))) {
-            if(c->ssl)
-                idx_cache_save(SSL_get_session(c->ssl),
-                    &c->connect_addr.addr[c->idx]);
+            if(c->ssl) {
+                SSL_SESSION *sess=SSL_get1_session(c->ssl);
+                if(sess) {
+                    idx_cache_save(sess, &c->connect_addr.addr[c->idx]);
+                    SSL_SESSION_free(sess);
+                }
+            }
             print_bound_address(c);
             fd=c->fd;
             c->fd=INVALID_SOCKET;
@@ -1430,6 +1572,7 @@ NOEXPORT void idx_cache_save(SSL_SESSION *sess, SOCKADDR_UNION *cur_addr) {
     SOCKADDR_UNION *old_addr, *new_addr;
     socklen_t len;
     char *addr_txt;
+    int ok;
 
     /* make a copy of the address, so it may work with delayed resolver */
     len=addr_len(cur_addr);
@@ -1442,12 +1585,12 @@ NOEXPORT void idx_cache_save(SSL_SESSION *sess, SOCKADDR_UNION *cur_addr) {
 
     CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_ADDR]);
     old_addr=SSL_SESSION_get_ex_data(sess, index_session_connect_address);
-    if(SSL_SESSION_set_ex_data(sess, index_session_connect_address, new_addr)) {
-        CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
+    ok=SSL_SESSION_set_ex_data(sess, index_session_connect_address, new_addr);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
+    if(ok) {
         str_free(old_addr); /* NULL pointers are ignored */
     } else { /* failed to store new_addr -> remove it */
         sslerror("SSL_SESSION_set_ex_data");
-        CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
         str_free(new_addr); /* NULL pointers are ignored */
     }
 }
@@ -1459,30 +1602,34 @@ NOEXPORT unsigned idx_cache_retrieve(CLI *c) {
     char *addr_txt;
 
     if(c->ssl && SSL_session_reused(c->ssl)) {
-        CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_ADDR]);
-        ptr=SSL_SESSION_get_ex_data(SSL_get_session(c->ssl),
-            index_session_connect_address);
-        if(ptr) {
-            len=addr_len(ptr);
-            memcpy(&addr, ptr, (size_t)len);
-            CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
-            /* address was copied, ptr itself is no longer valid */
-            for(i=0; i<c->connect_addr.num; ++i) {
-                if(addr_len(&c->connect_addr.addr[i])==len &&
-                        !memcmp(&c->connect_addr.addr[i],
-                            &addr, (size_t)len)) {
-                    addr_txt=s_ntop(&addr, len);
-                    s_log(LOG_INFO, "persistence: %s reused", addr_txt);
-                    str_free(addr_txt);
-                    return i;
+        SSL_SESSION *sess=SSL_get1_session(c->ssl);
+        if(sess) {
+            CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_ADDR]);
+            ptr=SSL_SESSION_get_ex_data(sess, index_session_connect_address);
+            if(ptr) {
+                len=addr_len(ptr);
+                memcpy(&addr, ptr, (size_t)len);
+                CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
+                SSL_SESSION_free(sess);
+                /* address was copied, ptr itself is no longer valid */
+                for(i=0; i<c->connect_addr.num; ++i) {
+                    if(addr_len(&c->connect_addr.addr[i])==len &&
+                            !memcmp(&c->connect_addr.addr[i],
+                                &addr, (size_t)len)) {
+                        addr_txt=s_ntop(&addr, len);
+                        s_log(LOG_INFO, "persistence: %s reused", addr_txt);
+                        str_free(addr_txt);
+                        return i;
+                    }
                 }
+                addr_txt=s_ntop(&addr, len);
+                s_log(LOG_INFO, "persistence: %s not available", addr_txt);
+                str_free(addr_txt);
+            } else {
+                CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
+                SSL_SESSION_free(sess);
+                s_log(LOG_NOTICE, "persistence: No cached address found");
             }
-            addr_txt=s_ntop(&addr, len);
-            s_log(LOG_INFO, "persistence: %s not available", addr_txt);
-            str_free(addr_txt);
-        } else {
-            CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
-            s_log(LOG_NOTICE, "persistence: No cached address found");
         }
     }
 
@@ -1623,12 +1770,19 @@ NOEXPORT int connect_init(CLI *c, int domain) {
 }
 
 NOEXPORT int redirect(CLI *c) {
+    SSL_SESSION *sess;
+    void *ex_data;
+
     if(!c->opt->redirect_addr.names)
         return 0; /* redirect not configured */
     if(!c->ssl)
         return 1; /* TLS not established -> always redirect */
-    return !SSL_SESSION_get_ex_data(SSL_get_session(c->ssl),
-        index_session_authenticated);
+    sess=SSL_get1_session(c->ssl);
+    if(!sess)
+        return 1; /* no TLS session -> always redirect */
+    ex_data=SSL_SESSION_get_ex_data(sess, index_session_authenticated);
+    SSL_SESSION_free(sess);
+    return ex_data == NULL;
 }
 
 NOEXPORT void print_bound_address(CLI *c) {
